#version 300 es
// Copyright 2022 The Android Open Source Project
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// ES 2 fragment shader that:
// 1. Samples from uTexSampler, copying from this texture to the current
//    output.
// 2. Applies a 4x4 RGB color matrix to change the pixel colors.
// 3. Transforms the optical colors to electrical colors using the SMPTE
//    170M OETF.

precision mediump float;
uniform sampler2D uTexSampler;
uniform mat4 uRgbMatrix;
in vec2 vTexSamplingCoord;
// C.java#ColorTransfer value.
// Only COLOR_TRANSFER_SDR and COLOR_TRANSFER_GAMMA_2_2 are allowed.
uniform int uOutputColorTransfer;
uniform vec2 vSurfaceSize;
out vec4 fragColor;
const float inverseGamma = 0.4500;

// Transforms a single channel from optical to electrical SDR using the SMPTE
// 170M OETF.
float smpte170mOetfSingleChannel(float opticalChannel) {
    // Specification:
    // https://www.itu.int/rec/R-REC-BT.1700-0-200502-I/en
    return opticalChannel < 0.018
    ? opticalChannel * 4.500
    : 1.099 * pow(opticalChannel, inverseGamma) - 0.099;
}

// Transforms optical SDR colors to electrical SDR using the SMPTE 170M OETF.
vec3 smpte170mOetf(vec3 opticalColor) {
    return vec3(smpte170mOetfSingleChannel(opticalColor.r),
    smpte170mOetfSingleChannel(opticalColor.g),
    smpte170mOetfSingleChannel(opticalColor.b));
}

// BT.709 gamma 2.2 OETF for one channel.
float gamma22OetfSingleChannel(highp float linearChannel) {
    // Reference:
    // https://developer.android.com/reference/android/hardware/DataSpace#TRANSFER_gamma22
    return pow(linearChannel, (1.0 / 2.2));
}

// BT.709 gamma 2.2 OETF.
vec3 gamma22Oetf(highp vec3 linearColor) {
    return vec3(gamma22OetfSingleChannel(linearColor.r),
    gamma22OetfSingleChannel(linearColor.g),
    gamma22OetfSingleChannel(linearColor.b));
}

// Applies the appropriate OETF to convert linear optical signals to nonlinear
// electrical signals. Input and output are both normalized to [0, 1].
highp vec3 applyOetf(highp vec3 linearColor) {
    // LINT.IfChange(color_transfer_oetf)
    const int COLOR_TRANSFER_SDR_VIDEO = 3;
    const int COLOR_TRANSFER_GAMMA_2_2 = 10;
    if (uOutputColorTransfer == COLOR_TRANSFER_SDR_VIDEO) {
        return smpte170mOetf(linearColor);
    } else if (uOutputColorTransfer == COLOR_TRANSFER_GAMMA_2_2) {
        return gamma22Oetf(linearColor);
    } else {
        // Output red as an obviously visible error.
        return vec3(1.0, 0.0, 0.0);
    }
}

void main() {
    // Scale constants
    vec4 scale = vec4(1. / vec2(textureSize(uTexSampler, 0)),
    vec2(textureSize(uTexSampler, 0)) / vSurfaceSize);
    // Source position in fractions of a texel
    vec2 src_pos = scale.zw * vec2(gl_FragCoord.x, gl_FragCoord.y);
    // Source bottom left texel centre
    vec2 src_centre = floor(src_pos - .5) + .5;
    // f is position. f.x runs left to right, y bottom to top, z right to left, w top to bottom
    vec4 f; f.zw = 1. - (f.xy = src_pos - src_centre);
    // Calculate weights in x and y in parallel.
    // These polynomials are piecewise approximation of Lanczos kernel
    vec4 l2_w0_o3 = ((1.5672 * f - 2.6445) * f + 0.0837) * f + 0.9976;
    vec4 l2_w1_o3 = ((-0.7389 * f + 1.3652) * f - 0.6295) * f - 0.0004;
    vec4 w1_2 = l2_w0_o3;
    vec2 w12 = w1_2.xy + w1_2.zw;
    vec4 wedge = l2_w1_o3.xyzw * w12.yxyx;
    // Calculate texture read positions. tc12 uses bilinear interpolation to do 4 reads in 1.
    vec2 tc12 = scale.xy * (src_centre + w1_2.zw / w12);
    vec2 tc0 = scale.xy * (src_centre - 1.);
    vec2 tc3 = scale.xy * (src_centre + 2.);
    // Sharpening adjustment
    float sum = wedge.x + wedge.y + wedge.z + wedge.w + w12.x * w12.y;

    // For Y the value is unsigned thus the range is [0.0, 1.0]
    wedge /= sum;

    vec3 col = vec3(
    texture(uTexSampler, vec2(tc12.x, tc0.y)).rgb * wedge.y +
    texture(uTexSampler, vec2(tc0.x, tc12.y)).rgb * wedge.x +
    texture(uTexSampler, tc12.xy).rgb * (w12.x * w12.y/sum) +
    texture(uTexSampler, vec2(tc3.x, tc12.y)).rgb * wedge.z +
    texture(uTexSampler, vec2(tc12.x, tc3.y)).rgb * wedge.w);


    vec4 transformedColors = uRgbMatrix * vec4(col, 1);

    fragColor = vec4(applyOetf(transformedColors.rgb), 1);
}
